from abc import abstractmethod

import psycopg2
from django.contrib.auth.models import Group
from django.contrib.gis.geos import GEOSException, MultiPolygon, Polygon, MultiPoint, MultiLineString
from django.core.exceptions import ObjectDoesNotExist
from django.core.files.uploadedfile import UploadedFile
from django.utils.timezone import make_aware

from konova.models import Geometry
from konova.settings import DEFAULT_GROUP
from user.models import User, UserActionLogEntry, UserAction, Team
from konova.tasks import celery_update_parcels


class BaseMigrater:
    options = None
    db_connection = None

    def __init__(self, options: dict):
        self.options = options
        self.connect_db()

    def __del__(self):
        if self.db_connection is not None:
            self.db_connection.close()

    def connect_db(self, db_name: str="ksp"):
        if self.options is None:
            return
        conn = psycopg2.connect(
            host=self.options["host"],
            database=db_name,
            user=self.options["db_user"],
            password=self.options["db_pw"],
        )
        self.db_connection = conn

    @abstractmethod
    def migrate(self):
        raise NotImplementedError("Must be implemented in subclass")

    def _migrate_geometry(self, instance, db_result: tuple):
        try:
            # db_result[2] == polygon, db_result[3] == line, db_result[4] == point
            if db_result[2] is not None:
                db_result_geom = MultiPolygon.from_ewkt(db_result[2])
            elif db_result[3] is not None:
                db_result_geom = MultiLineString.from_ewkt(db_result[3])
                db_result_geom = db_result_geom.buffer(0.00001, 1)
                if isinstance(db_result_geom, Polygon):
                    db_result_geom = MultiPolygon(db_result_geom)
            elif db_result[4] is not None:
                db_result_geom = MultiPoint.from_ewkt(db_result[4])
                db_result_geom = db_result_geom.buffer(0.00001, 1)
                if isinstance(db_result_geom, Polygon):
                    db_result_geom = MultiPolygon(db_result_geom)
            else:
                db_result_geom = None
        except GEOSException:
            print(f"Malicious geometry on {db_result}")
            return
        geom = instance.geometry or Geometry()
        geom.geom = db_result_geom
        geom.save()
        celery_update_parcels.delay(geom.id)

        instance.geometry = geom
        return instance

    def _migrate_documents(self, instance, document_cls, db_result):
        db_result_identifier = f"'{db_result[0]}'"
        tmp_cursor = self.db_connection.cursor()
        tmp_cursor.execute(
            'select '
                'doc.pfad, '
                'doc."Bemerkung", '
                'doc."Datum" '
            'from "OBJ_MASTER" om '
                'left join "LINFOS" linf on om."GISPADID"=linf."GISPADID" '
                'left join "Objektphotos" doc on linf."PKEY"=doc."FKEY" '
            'where '
                f'om."KENNUNG"={db_result_identifier} and '
                'doc.pfad is not null '
        )
        doc_results = tmp_cursor.fetchall()
        if len(doc_results) > 0:
            for doc_result in doc_results:
                doc_path = doc_result[0]
                doc_comment = doc_result[1]
                doc_date = doc_result[2]
                try:
                    with open(doc_path, encoding="latin1") as file:
                        file = UploadedFile(file)
                        doc_title = "Migrierte Datei"
                        doc_date = doc_date or "1970-01-01"
                        doc_exists = document_cls.objects.filter(
                            instance=instance,
                            title=doc_title,
                            comment=doc_comment,
                            date_of_creation=doc_date
                        ).exists()
                        if doc_exists:
                            continue
                        doc = document_cls.objects.create(
                            title=doc_title,
                            comment=doc_comment,
                            file=file,
                            date_of_creation=doc_date,
                            instance=instance,
                        )
                except FileNotFoundError:
                    print(f"------ !!! File not found: {doc_path}")
        tmp_cursor.close()
        return instance

    def _migrate_log(self, instance, db_result: tuple):
        identifier = f"'{db_result[0]}'"
        tmp_cursor = self.db_connection.cursor()
        tmp_cursor.execute(
            'select '
                'p.bemerkung1, '
                'p.geaendertam, '
                'p.geaendertvon '
            'from "OBJ_MASTER" om '
                'join protokoll p on om."GISPADID"=p."FKEY" '
            'where '
                f'om."KENNUNG"={identifier}'
        )
        fetch_results = tmp_cursor.fetchall()

        if len(fetch_results) == 0:
            # For whatever reason, there is another logging table. So if there are no entries on "protokoll", we
            # need to take a look on "log" as a fallback
            tmp_cursor.execute(
                'select '
                    'l.nachricht, '
                    'l.erstelltam, '
                    'l.erstelltvon '
                'from "OBJ_MASTER" om '
                    'join log l on om."GISPADID"=l.gispadid::Integer '
                'where '
                    f'om."KENNUNG"={identifier}'
            )
            fetch_results = tmp_cursor.fetchall()

        migrated_identifier = "(migriert)"
        for result in fetch_results:
            comment = f"{result[0]} {migrated_identifier}"
            timestamp = make_aware(result[1])
            user_name = result[2]
            user = self._get_migrate_user(user_name)

            try:
                action = instance.log.get(
                    user=user,
                    timestamp=timestamp,
                    action=UserAction.EDITED,
                    comment=comment
                )
            except ObjectDoesNotExist:
                action = UserActionLogEntry(
                    user=user,
                    timestamp=timestamp,
                    action=UserAction.EDITED,
                    comment=comment
                )
                action.save()
                action.timestamp = timestamp
                action.save()
                instance.log.add(action)

        first_entry = instance.log.order_by("timestamp").first()
        last_entry = instance.log.order_by("-timestamp").first()

        if first_entry is not None and instance.created is None:
            first_entry.action = UserAction.CREATED
            first_entry.save()
            instance.created = first_entry
        if last_entry is not None and instance.modified != last_entry:
            instance.modified = last_entry

        tmp_cursor.close()
        return instance

    def _get_team(self, team_name: str):
        """ Returns a team from given name

        """
        if len(team_name) == 0:
            return None

        name = f"Team {team_name}"
        description = f"Automatisch erzeugtes Team für {team_name}"
        team = Team.objects.get_or_create(
            name=name,
            description=description
        )[0]
        return team

    def _get_team_from_responsible_code(self, responsible_code, prefix: str = "Team"):
        """ Returns a team from given responsible code

        """
        name = f"{prefix} {responsible_code.long_name}"
        if responsible_code.parent is not None:
            name += f", {responsible_code.parent.long_name}"

        description = f"Automatisch erzeugtes Team für {str(responsible_code)}"
        team = Team.objects.get_or_create(
            name=name,
            description=description
        )[0]
        return team

    def _migrate_responsible_code_to_team(self, instance, responsible_code, prefix: str = "Team"):
        team = self._get_team_from_responsible_code(responsible_code, prefix)
        instance.share_with_team(team)

        return instance

    def _migrate_recorded(self, instance, db_result):
        raise NotImplementedError("Must be implemented in specific subtype classes!")

    def _get_migrate_user(self, username):
        """ Returns a migrated user

        """
        user = User.objects.get_or_create(
            username=username,
        )
        is_new = user[1]
        user = user[0]
        if is_new:
            user.first_name = "MIGRIERT"
            user.last_name = "MIGRIERT"
            user.save()

        # Make sure user has at least the default group set
        default_group = Group.objects.get(name=DEFAULT_GROUP)
        user.groups.add(default_group)
        return user